/*
 * Created on Dec 12, 2003 by sviglas
 *
 * Modified on Dec 24, 2008 by sviglas
 *
 * This is part of the attica project.  Any subsequent modification
 * of the file should retain this disclaimer.
 * 
 * University of Edinburgh, School of Informatics
 */
package org.dejave.attica.engine.operators;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import java.io.IOException;

import org.dejave.attica.model.Relation;

import org.dejave.attica.engine.predicates.Predicate;
import org.dejave.attica.engine.predicates.PredicateEvaluator;
import org.dejave.attica.engine.predicates.PredicateTupleInserter;

import org.dejave.attica.storage.IntermediateTupleIdentifier;
import org.dejave.attica.storage.RelationIOManager;
import org.dejave.attica.storage.StorageManager;
import org.dejave.attica.storage.StorageManagerException;
import org.dejave.attica.storage.Tuple;

import org.dejave.attica.storage.FileUtil;

/**
 * NestedLoopsJoin: Implements a (blocking) nested loops join.
 *
 * @author sviglas
 */
public class NestedLoopsJoin extends PhysicalJoin {

    /** The name of the temporary file for the left input. */
    private String leftFile;
    
    /** The name of the temporary file for the right input. */
    private String rightFile;
	
    /** The name of the temporary file for the output. */
    private String outputFile;
	
    /** The relation manager used for I/O. */
    private RelationIOManager outputMan;

    /** The iterator over the output file. */
    private Iterator<Tuple> outputTuples;

    /** Reusable output list. */
    private List<Tuple> returnList;
	
    /**
     * Constructs a new nested loops join operator.
     * 
     * @param left the left input operator.
     * @param right the right input operator.
     * @param sm the storage manager.
     * @param predicate the predicate evaluated by this join operator.
     * @throws EngineException thrown whenever the operator cannot be 
     * properly constructed.
     */
    public NestedLoopsJoin (Operator left, Operator right,
                            StorageManager sm, Predicate predicate) 
	throws EngineException {
        
        super(left, right, sm, predicate);
        try {
            leftFile = FileUtil.createTempFileName();
            sm.createFile(leftFile);
            rightFile = FileUtil.createTempFileName();
            sm.createFile(rightFile);
            returnList = new ArrayList<Tuple>();            
        }
        catch (StorageManagerException sme) {
            throw new EngineException("Could not instantiate "
                                      + "nested-loops join", sme);
        }
    } // NestedLoopsJoin()

    
    /**
     * Sets up a nested loops join operator.
     * 
     * @throws EngineException thrown whenever there is something
     * wrong with setting this operator up.
     */
    @Override
    protected void setup() throws EngineException {
        try {
            // this is a blocking operator -- store the input
            // in temporary files
            
            // store the left input
            Relation leftRel = getInputOperator(LEFT).getOutputRelation();
            RelationIOManager leftMan =
                new RelationIOManager(getStorageManager(), leftRel, leftFile);
            boolean done = false;
            while (! done) {
                Tuple tuple = getInputOperator(LEFT).getNext();
                if (tuple != null) {
                    done = (tuple instanceof EndOfStreamTuple);
                    if (! done) leftMan.insertTuple(tuple);
                }
            }
            
            // store the right input
            Relation rightRel = getInputOperator(RIGHT).getOutputRelation();
            RelationIOManager rightMan = 
                new RelationIOManager(getStorageManager(), rightRel, rightFile);
            done = false;
            while (! done) {
                Tuple tuple = getInputOperator(RIGHT).getNext();
                if (tuple != null) {
                    done = (tuple instanceof EndOfStreamTuple);
                    if (! done) rightMan.insertTuple(tuple);
                }
            }
            
            // the inputs are now stored -- perform the nested loops join
            outputFile = FileUtil.createTempFileName();
            getStorageManager().createFile(outputFile);
            outputMan = new RelationIOManager(getStorageManager(), 
                                              getOutputRelation(), 
                                              outputFile);
            for (Tuple leftTuple : leftMan.tuples()) {
                for (Tuple rightTuple : rightMan.tuples()) {
                    PredicateTupleInserter.insertTuples(leftTuple, 
                                                        rightTuple,
                                                        getPredicate());
                    if (PredicateEvaluator.evaluate(getPredicate())) {
                        // the predicate is true -- store the new tuple
                        Tuple newTuple = combineTuples(leftTuple, rightTuple);
                        outputMan.insertTuple(newTuple);
                    }
                }
            }
            
            //outputMan = new RelationIOManager(getStorageManager(),
            //                                  getOutputRelation(),
            //                                  outputFile);
            // now open the iterator over the output
            outputTuples = outputMan.tuples().iterator();
        }
        catch (IOException ioe) {
            throw new EngineException("Could not create page/tuple iterators.",
                                      ioe);
        }
        catch (StorageManagerException sme) {
            throw new EngineException("Could not store intermediate relations " + 
                                      "to files.", sme);
        }
    } // setup()

    
    /**
     * Cleanup after the join.
     * 
     * @throws EngineException whenever the operator cannot clean up
     * after itself.
     */
    @Override
    protected void cleanup() throws EngineException {
        try {
            getStorageManager().deleteFile(leftFile);
            getStorageManager().deleteFile(rightFile);
            getStorageManager().deleteFile(outputFile);
        }
        catch (StorageManagerException sme) {
            throw new EngineException("Could not clean up final output", sme);
        }
    } // cleanup()
    
    /**
     * Inner method to propagate a tuple.
     * 
     * @return an array of resulting tuples.
     * @throws EngineException thrown whenever there is an error in
     * execution.
     */
    @Override
    protected List<Tuple> innerGetNext () throws EngineException {
        try {
            returnList.clear();
            if (outputTuples.hasNext()) returnList.add(outputTuples.next());
            else returnList.add(new EndOfStreamTuple());
            return returnList;
        }
        catch (Exception sme) {
            throw new EngineException("Could not read tuples "
                                      + "from intermediate file.", sme);
        }
    } // innerGetNext()
        

    @Override
    protected List<Tuple> innerProcessTuple(Tuple tuple, int inOp)
	throws EngineException {
        
        return new ArrayList<Tuple>();
    }  // innerProcessTuple()

    
    /**
     * Textual representation.
     */
    @Override
    protected String toStringSingle() {
        return "nlj <" + getPredicate() + ">";
    } // toStringSingle()

} // NestedLoopsJoin
